use std::io;
use std::os::unix::net::UnixStream;
use std::io::{Read, Write};
use clap::Parser;
use std::num::ParseIntError;
use std::process::{Command, Stdio};

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// Path to the unix socket for the monitor
    #[arg(long)]
    monitor_socket_path: String,

    /// Path to the unix socket for the uart
    #[arg(long)]
    uart_socket_path: String,

    /// RAM file
    #[arg(long)]
    ram_file: String,
}

fn parse_hex(s: &str) -> Result<u64, ParseIntError> {
    u64::from_str_radix(s.trim_start_matches("0x"), 16)
}

fn print_menu() {
    println!("Menu:");
    println!("1. Read data");
    println!("2. Write data");
    println!("3. Listen UART");
    println!("4. Listen UART until string");
    println!("5. Watch physical page");
    println!("6. Is watched page hit?");
    println!("7. Resume execution");
    println!("8. Clear watched page");
}

fn read_data(ram_size: u64, mut socket: &mut UnixStream) -> Result<String, String> {
    println!("Physical Adddress (hex):");
    let pa : u64 = {
        let mut pa_str = String::new();
        io::stdin().read_line(&mut pa_str).expect("Failed to read line");
        parse_hex(pa_str.trim()).unwrap()
    };

    if pa >= ram_size {
        return Err("PA beyond RAM size".to_string());
    }

    println!("Num bytes:");
    let num_bytes : u64 = {
        let mut num_str = String::new();
        io::stdin().read_line(&mut num_str).expect("Failed to read line");
        num_str.trim().parse().unwrap()
    };

    if num_bytes % 8 != 0 {
        return Err("Num bytes must be 8-byte divisible".to_string());
    }

    if pa + num_bytes > ram_size {
        return Err("PA beyond RAM size".to_string());
    }

    let mut hex_bytes = String::new();
    for index in 0..(num_bytes / 8) {
        let command = format!("xp /xg 0x{:x}", pa + index * 8u64);
        send_qemu_monitor_command(&mut socket, &command);
        let _ = read_line(&mut socket);
        let res = read_line(&mut socket);
        let value = parse_hex(res.split(": ").collect::<Vec<&str>>()[1].trim()).unwrap();
        let hex_string = format!("{:016x}", value.swap_bytes());
        hex_bytes.push_str(&hex_string);
    }

    Ok(hex_bytes)
}

fn write_data(ram_size: u64, mut socket: &mut UnixStream) -> Result<String, String> {
    println!("Physical Adddress (hex):");
    let pa : u64 = {
        let mut pa_str = String::new();
        io::stdin().read_line(&mut pa_str).expect("Failed to read line");
        parse_hex(pa_str.trim()).unwrap()
    };

    if pa >= ram_size {
        return Err("PA beyond RAM size".to_string());
    }

    println!("Num bytes:");
    let num_bytes : u64 = {
        let mut num_str = String::new();
        io::stdin().read_line(&mut num_str).expect("Failed to read line");
        num_str.trim().parse().unwrap()
    };

    if num_bytes % 8 != 0 {
        return Err("Num bytes must be 8-byte divisible".to_string());
    }

    if pa + num_bytes > ram_size {
        return Err("PA beyond RAM size".to_string());
    }

    println!("Data: ");
    let mut data = vec![0u8; num_bytes as usize];
    if let Err(_) = io::stdin().read_exact(&mut data) {
        return Err(String::from("failed to read data"));
    }

    for (chunk_index, chunk) in data.chunks(8).enumerate() {
        let block_value = unsafe { *(chunk.as_ptr() as *const u64) };
        let command = format!("set_phys_8 0x{:x} 0x{:x}", pa + (chunk_index as u64) * 8u64, block_value);
        send_qemu_monitor_command(&mut socket, &command);
    }

    Ok(String::from("Success"))
}

fn watch_physical_page(mut monitor_socket: &mut UnixStream) -> Result<String, String> {
    println!("Physical Adddress (hex):");
    let pa : u64 = {
        let mut pa_str = String::new();
        io::stdin().read_line(&mut pa_str).expect("Failed to read line");
        parse_hex(pa_str.trim()).unwrap()
    };

    let command = format!("watch_physical_page 0x{:x}", pa);
    send_qemu_monitor_command(&mut monitor_socket, &command);

    Ok(String::from("Success"))
}

fn is_watched_page_hit(mut monitor_socket: &mut UnixStream) -> Result<String, String> {
    send_qemu_monitor_command(&mut monitor_socket, "watched_page_hit");
    let _ = read_until_string(monitor_socket, "Page hit? ");
    let res = read_until_string(monitor_socket, "\n");
    let res = &res[..res.len() - 1];
    println!("{}", res);

    Ok(String::from("Success"))
}

fn resume_execution(mut monitor_socket: &mut UnixStream) -> Result<String, String> {
    send_qemu_monitor_command(&mut monitor_socket, "resume_all_vcpus");
    Ok(String::from("Success"))
}

fn clear_watched_page(mut monitor_socket: &mut UnixStream) -> Result<String, String> {
    send_qemu_monitor_command(&mut monitor_socket, "clear_watched_page");
    Ok(String::from("Success"))
}

fn listen_uart(socket: &mut UnixStream) -> Result<String, String> {
    println!("Num bytes:");
    let num_bytes : u64 = {
        let mut num_str = String::new();
        io::stdin().read_line(&mut num_str).expect("Failed to read line");
        num_str.trim().parse().unwrap()
    };

    for _ in 0u64..num_bytes {
        let mut b = [0u8];
        if let Err(_) = socket.read_exact(&mut b) {
            return Err(String::from("Failed to read data"));
        }
        io::stdout().write(&b).unwrap();
        io::stdout().flush().unwrap();
    }
    Ok(String::from("Success"))
}

fn listen_uart_until_string(socket: &mut UnixStream) -> Result<String, String> {
    println!("Num bytes:");
    let num_bytes : u64 = {
        let mut num_str = String::new();
        io::stdin().read_line(&mut num_str).expect("Failed to read line");
        num_str.trim().parse().unwrap()
    };
    println!("Pattern:");
    let mut pattern = vec![0u8; num_bytes as usize];
    if let Err(_) = io::stdin().read_exact(&mut pattern) {
        return Err(String::from("failed to read data"));
    }

    let mut data = Vec::new();
    loop {
        let mut b = [0u8];
        if let Err(_) = socket.read_exact(&mut b) {
            return Err(String::from("Failed to read data"));
        }
        data.push(b[0]);
        if data.ends_with(pattern.as_slice()) {
            break;
        }
    }
    io::stdout().write(data.as_slice()).unwrap();
    io::stdout().flush().unwrap();
    Ok(String::from("Success"))
}

fn read_line(socket: &mut UnixStream) -> String {
    let result = read_until_string(socket, "\n");
    result[..result.len() - 1].to_string()
}

fn read_until_string(socket: &mut UnixStream, pattern: &str) -> String {
    let mut data = String::new();
    loop {
        let mut buf = [0u8];
        if let Ok(()) = socket.read_exact(&mut buf) {
            let ch = buf[0] as char;
            data.push(ch);
            if data.ends_with(&pattern[..]) {
                return data;
            }
        }
    }
}

fn send_qemu_monitor_command(socket: &mut UnixStream, command: &str) {
    let _ = read_until_string(socket, "(qemu) ");
    socket.write_all(command.as_bytes()).unwrap();
    socket.write_all(b"\n").unwrap();
}

fn main() {
    // PLEASE NOTE:
    // The default UART unix socket has a buffer limit that is very easy to hit with the
    // excessive logging of OVMF (+enabled logging) and Linux. When the buffer becomes full,
    // this will block subsequent execution of the VM.
    //
    // A way to WAR the issue is to read some data from UART to make space for more
    // writes.

    // Dump the buffer size (intended for debugging):
    // {
    //     println!("Unix socket buffering size: ");
    //     Command::new("cat")
    //         .args(["/proc/sys/net/core/wmem_default"])
    //         .stdout(Stdio::inherit())
    //         .stderr(Stdio::inherit())
    //         .spawn().expect("Failed to dump unix socket limit")
    //         .wait().expect("Unexpected process exit");
    // }

    let args = Args::parse();

    let ram_size = 256u64 * 1024 * 1024;

    // Launch VM and wait a tiny bit before the sockets are opened
    let mut process = Command::new("./launch_vm.sh")
        .args([
            "--monitor_unix_path", &args.monitor_socket_path,
            "--uart_unix_path", &args.uart_socket_path,
            "--ram_size", &format!("{}", ram_size),
            "--ram_file", &args.ram_file
        ])
        .stdin(Stdio::null())
        // TODO:
        // .stdout(Stdio::null())
        // .stderr(Stdio::null())
        .stdout(Stdio::inherit())
        .stderr(Stdio::inherit())
        .spawn()
        .expect("Failed to launch VM");
    std::thread::sleep(core::time::Duration::from_secs(1));

    if let Ok(Some(ret)) = process.try_wait() {
        let output = process.wait_with_output().unwrap();
        println!("Process exitted unexpectedly: {}.\nStdout {}.\nStderr {}",
            ret, String::from_utf8(output.stdout).unwrap(), String::from_utf8(output.stderr).unwrap());
        return;
    }

    // Connect to sockets
    let mut monitor_socket = UnixStream::connect(&args.monitor_socket_path).unwrap();
    let mut uart_socket = UnixStream::connect(&args.uart_socket_path).unwrap();

    println!("Another one of these Confidential Computing VM challenges.");
    println!("Here and ready for hacking!");

    loop {
        print_menu();
        let mut line = String::new();
        io::stdin().read_line(&mut line).expect("Failed to read line");
        match line.trim().parse() {
            Ok(1) => {
                match read_data(ram_size, &mut monitor_socket) {
                    Ok(data) => println!("Data: {}", data),
                    Err(msg) => println!("Error: {}", msg),
                }
            },
            Ok(2) => {
                match write_data(ram_size, &mut monitor_socket) {
                    Ok(msg) => println!("{}", msg),
                    Err(msg) => println!("Error: {}", msg),
                }
            }
            Ok(3) => {
                match listen_uart(&mut uart_socket) {
                    Ok(_) => {},
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(4) => {
                match listen_uart_until_string(&mut uart_socket) {
                    Ok(_) => continue,
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(5) => {
                match watch_physical_page(&mut monitor_socket) {
                    Ok(_) => {},
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(6) => {
                match is_watched_page_hit(&mut monitor_socket) {
                    Ok(_) => {},
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(7) => {
                match resume_execution(&mut monitor_socket) {
                    Ok(_) => {},
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(8) => {
                match clear_watched_page(&mut monitor_socket) {
                    Ok(_) => {},
                    Err(err) => println!("Error: {}", err),
                }
            }
            Ok(_) => {
                println!("Unknown input");
            }
            Err(_) => {
                println!("Invalid input");
            }
        }
        io::stdout().flush().unwrap();
    }
}
